// EasyBEPU a basic example of BEPU with XNA

// made from XNA fork version of BEPUPhysicsDemo sample code
// with a lot of extra stuff and features removed to be as short as possible
// and having only basic XNA drawing features with basic BEPU features
// Feel free to replace the walldemo used here by another one of your choice and add not included feature like
// bounding box or joint display, the car model, ...

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

using BEPUphysics;
using BEPUphysics.Threading;
using BEPUphysicsDrawer.Models;
using BEPUphysics.Entities.Prefabs;
using BEPUphysics.Entities;

namespace EasyBEPU
{
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;

        private ParallelLooper parallelLooper;

        public Camera Camera;
        public ModelDrawer ModelDrawer;

        //Input
        public KeyboardState KeyboardInput;
        public KeyboardState PreviousKeyboardInput;
        public MouseState MouseInput;
        public MouseState PreviousMouseInput;

        // fun stuff
        public Entity kapow;
        //public Explosion kapowMaker;

        Box ground;
        public Space Space;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            parallelLooper = new ParallelLooper();

            if (Environment.ProcessorCount > 1)
            {
                for (int i = 0; i < Environment.ProcessorCount; i++)
                {
                    parallelLooper.AddThread();
                }
            }

            Space = new Space(parallelLooper);

            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;

            Camera = new Camera(BEPUutilities.Vector3.Zero, 0, 0, BEPUutilities.Matrix.CreatePerspectiveFieldOfViewRH(BEPUutilities.MathHelper.PiOver4, graphics.PreferredBackBufferWidth / (float)graphics.PreferredBackBufferHeight, .1f, 10000));

            Camera.LockedUp = BEPUutilities.Vector3.Up;
            Camera.ViewDirection = new BEPUutilities.Vector3(0, 0, -1);

            kapow = new Sphere(new BEPUutilities.Vector3(11000, 0, 0), .6f, 20);
            //kapowMaker = new Explosion(Vector3.Zero, 400, 15, Space);
            Space.Add(kapow);

            // try disabling this one... remember Matrix ? :)
            Space.ForceUpdater.Gravity = new BEPUutilities.Vector3(0, -9.81f, 0f);

        }


        protected override void Initialize()
        {
            ModelDrawer = new InstancedModelDrawer(this);

            base.Initialize();
        }

        protected override void LoadContent()
        {
            IsFixedTimeStep = false;

            //Clear out any old rendering stuff.
            ModelDrawer.Clear();

            // Here is the demo !
            WallDemo();

            // Quite slow here with console output for some reason
            foreach (Entity e in Space.Entities)
            {
                //Console.WriteLine("Entity tag:[" + e.ToString() + "] checked for model drawer");
                ModelDrawer.Add(e);
                //Console.WriteLine("Entity added to ModelDrawer");
            }

            // probably not needed here
            //for (int i = 0; i < Space.Solver.SolverUpdateables.Count; i++)
            //{
            //    //Add the solver updateable and match up the activity setting.
            //    LineDisplayObjectBase objectAdded = ConstraintDrawer.Add(Space.Solver.SolverUpdateables[i]);
            //    if (objectAdded != null)
            //        objectAdded.IsDrawing = Space.Solver.SolverUpdateables[i].IsActive;
            //}

            Mouse.SetPosition(200, 200);

            GC.Collect();
        }

        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        protected override void Update(GameTime gameTime)
        {
            PreviousKeyboardInput = KeyboardInput;
            KeyboardInput = Keyboard.GetState();
            PreviousMouseInput = MouseInput;
            MouseInput = Mouse.GetState();

            var dt = (float)gameTime.ElapsedGameTime.TotalSeconds;

            if (KeyboardInput.IsKeyDown(Keys.Escape))
            {
                parallelLooper.Dispose();

                Exit();
            }

            if (!IsMouseVisible)
                Mouse.SetPosition(200, 200);

            Camera.Yaw((200 - MouseInput.X) * dt * 0.012f);
            Camera.Pitch((200 - MouseInput.Y) * dt * 0.012f);

            float Speed = 10f;
            float distance = Speed * dt;

            if (KeyboardInput.IsKeyDown(Keys.E))
                Camera.MoveForward(distance);
            if (KeyboardInput.IsKeyDown(Keys.D))
                Camera.MoveForward(-distance);
            if (KeyboardInput.IsKeyDown(Keys.S))
                Camera.MoveRight(-distance);
            if (KeyboardInput.IsKeyDown(Keys.F))
                Camera.MoveRight(distance);
            if (KeyboardInput.IsKeyDown(Keys.A))
                Camera.MoveUp(distance);
            if (KeyboardInput.IsKeyDown(Keys.Z))
                Camera.MoveUp(-distance);

            // A very basic "shaking" of the board, just for fun
            if (KeyboardInput.IsKeyDown(Keys.P) && PreviousKeyboardInput.IsKeyUp(Keys.P))
            {
                //ground.ApplyImpulse(BEPUutilities.Vector3.Zero, new BEPUutilities.Vector3(0, 5, 0));
                ground.LinearVelocity = new BEPUutilities.Vector3(0, 5, 0);

            }
            if (KeyboardInput.IsKeyDown(Keys.O) && PreviousKeyboardInput.IsKeyUp(Keys.O))
            {
                ground.LinearVelocity = BEPUutilities.Vector3.Zero;

            }

            if (MouseInput.LeftButton == ButtonState.Pressed)
            {
                kapow.Position = Camera.Position + Camera.WorldMatrix.Forward;
                kapow.AngularVelocity = BEPUutilities.Vector3.Zero;
                kapow.LinearVelocity = Camera.WorldMatrix.Forward * 30; // too high speed makes collision undetected
            }

            if (MouseInput.XButton1 == Microsoft.Xna.Framework.Input.ButtonState.Pressed)
            {
                //Interpolation isn't used in the demos by default, so passing in a really short time adds a lot of time between discretely visible time steps.
                //Using a Space.TimeStepSettings.TimeStepDuration of 1/60f (the default), this will perform one time step every 20 frames (about three per second at the usual game update rate).
                //This can make it easier to examine behavior frame-by-frame.
                Space.Update(1 / 1200f);
            }
            else
            {
                Space.Update();
            }

            ModelDrawer.Update();

            PreviousKeyboardInput = KeyboardInput;
            PreviousMouseInput = MouseInput;

            base.Update(gameTime);
        }

        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(new Color(.41f, .41f, .45f, 1));

            var viewMatrix = Camera.ViewMatrix;
            var projectionMatrix = Camera.ProjectionMatrix;

            ModelDrawer.Draw(viewMatrix, projectionMatrix);

            base.Draw(gameTime);
        }



        // demo code, example of the wall demo, one of the smallest demo, as it doesn't have any dependancy
        public void WallDemo()
        {
            int width = 10;
            int height = 10;
            float blockWidth = 2f;
            float blockHeight = 1f;
            float blockLength = 1f;

            for (int i = 0; i < width; i++)
            {
                for (int j = 0; j < height; j++)
                {
                    var toAdd =
                        new Box(
                            new BEPUutilities.Vector3(
                                i * blockWidth + .5f * blockWidth * (j % 2) - width * blockWidth * .5f,
                                blockHeight * .5f + j * (blockHeight),
                                0),
                            blockWidth, blockHeight, blockLength, 10);
                    Space.Add(toAdd);
                }
            }

            //Box 
            ground = new Box(new BEPUutilities.Vector3(0, -.5f, 0), 50, 1, 50);
            Space.Add(ground);
            Camera.Position = new BEPUutilities.Vector3(0, 6, 15);
        }
    }
}
